"""
Provides tools for generation of auto-differentiable operations.

See https://github.com/theHamsta/pystencils_autodiff

Installation:

.. code-block:: bash
    pip install pystencils-autodiff
"""

raise NotImplementedError('pystencils-autodiff is not installed. Run `pip install pystencils-autodiff`')
